#include <stdio.h>
#include "ext.hpp"

const char* direction (int &dir) {
	switch (dir) {
		case north: return "North";
		case east: return "East";
		case south: return "South";
		case west: return "West";
		default: return "nowhere";
	}
}

const char* alarm (int &temperature) {
	switch (temperature) {
		case safe: return "safe";
		case hot: return "hot";
		case burning: return "burning";
		default: return "weird";
	}
}

int rotate_dir (int &dir) {
	return (dir << 1) % dir_modulus;
}

int delta_x (int &dir) {
	switch (dir) {
		case east: return 1;
		case west: return -1;
		default: return 0;
	}
}

int delta_y (int &dir) {
	switch (dir) {
		case north: return -1;
		case south: return 1;
		default: return 0;
	}
}

/* Square row printing */

void print_square_row (int size, Square sq[], int target, int robot, int dir) {
	for (int i = 0; i < size; i++) 
		printf (".%c.", (sq[i].border & north) ? '-' : ' ');
	printf ("\n");
	for (int i = 0; i < size; i++) {
		char cell = ' ';
		if (robot == i + 1) {
			switch (dir) {
				case north: 
					cell = '^';
					break;
				case east: 
					cell = '>';
					break;
				case south: 
					cell = 'v';
					break;
				case west: 
					cell = '<';
					break;
				default:
					cell = ' ';
					break;
			}
		}

		switch (sq[i].temperature) {
			case safe: 
				cell = (target == i + 1) ? 'T' : cell; 
				break;
			case hot:
				cell = (target == i + 1) ? 'H' : 'h';
				break;
			case burning:
				cell = (target == i + 1) ? 'B' : 'b';
				break;
			default:
				cell = (target == i + 1) ? 'X' : 'x';
				break;
		}

		printf ("%c%c%c", 
			(sq[i].border & west) ? '|' : ' ', 
			cell,
			(sq[i].border & east) ? '|' : ' ');
	}	
	printf ("\n");
	for (int i = 0; i < size; i++) 
		printf (".%c.", (sq[i].border & south) ? '_' : ' ');
}

/* Field class implementation */

void Field::print_field (int robot_x, int robot_y, int dir) {
	for (int i = 0; i < FIELD_SIZE; i++)
		printf("*****");

	for (int i = 0; i < FIELD_SIZE; i++) {
		printf("\n");
		print_square_row (FIELD_SIZE, field[i], 
			(target_y == i + 1) ? target_x : 0, 
			(robot_y == i + 1) ? robot_x : 0, dir);
	}

	printf("\n");
	for (int i = 0; i < FIELD_SIZE; i++)
		printf("*****");
	printf("\n");
}

void Field::init_square (int i, int j, temperature_type t) {
	Square &sq = this->field[j][i];
	sq.timer = 0;
	sq.temperature = t;
	sq.volatility = 0;
	sq.border = NO_BORDER;
	sq.border |= (i == 0) ? west : 0;
	sq.border |= (i == FIELD_SIZE - 1) ? east : 0;
	sq.border |= (j == 0) ? north : 0;
	sq.border |= (j == FIELD_SIZE - 1) ? south : 0;
}

void Field::evolve_square (int i, int j) {
	Square &sq = field[j][i];
	switch (sq.temperature) {
		case safe:
			if (sq.volatility == 1 &&
					((i > 0 && field[j][i-1].temperature == burning) ||
			 		 (i < FIELD_SIZE - 1 && field[j][i+1].temperature == burning) ||
			 		 (j > 0 && field[j-1][i].temperature == burning) ||
			 		 (j < FIELD_SIZE - 1 && field[j+1][i].temperature == burning))) {
				sq.timer = 0;
				sq.temperature = hot;
			}
			break;
		case hot:
			if (sq.timer >= delay) {
				sq.timer = 0;
				sq.temperature = (sq.volatility == 1) ? burning : safe;
			}
			break;
		case burning:
			if (sq.timer >= 3*delay) {
				sq.timer = 0;
				sq.temperature = hot;
			}
			break;
	}
	
	sq.volatility = 1 - sq.volatility;
	sq.timer++;
}

void Field::init_field (int t_x, int t_y, int delay) {
	delay = delay;
	target_x = t_x;
	target_y = t_y;

	for (int i = 0; i < FIELD_SIZE; i++)
		for (int j = 0; j < FIELD_SIZE; j++) {
			temperature_type temp = 
				((i == 2 && j == 2)) ?// || ((i == 1 || i == 3) && (j == 3 || j == 1))) ? 
				burning : safe;
			init_square(i, j, temp);
		}
}

void Field::evolve_field() {
	for (int i = 0; i < FIELD_SIZE; i++)
		for (int j = 0; j < FIELD_SIZE; j++)
			evolve_square(i,j);	
}

int Field::detect_target(int x, int y) {
	return (x == this->target_x && y == this->target_y) ? 1 : 0;
}

int Field::detect_borders(int x, int y) {
	return this->field[y-1][x-1].border;
}

int Field::temperature_in_front(int x, int y, int dir) {
	switch (dir) {
		case north: return (y > 1) ? this->field[y-2][x-1].temperature : burning;
		case east:  return (x < FIELD_SIZE) ? this->field[y-1][x].temperature : burning;
		case south: return (y < FIELD_SIZE) ? this->field[y][x-1].temperature : burning;
		case west:  return (x > 1) ? this->field[y-1][x-2].temperature : burning;
		default: return burning;
	}	
}

void init_field (Field &field, int t_x, int t_y, int delay) {
	field.init_field (t_x, t_y, delay);
}

int detect_target (Field &field, int &x, int &y) {
	return field.detect_target(x, y);
}

int detect_borders(Field &field, int &x, int &y) {
	return field.detect_borders(x, y);
}

int temperature_in_front(Field &field, int &x, int &y, int &dir) {
	return field.temperature_in_front(x, y, dir);
}

/* Global variable field */

Field field;

void init_field () {
	init_field (field, TARGET_X, TARGET_Y, DELAY);
}

int detect_target (int &x, int &y) {
	return detect_target (field, x, y);
}

int detect_borders(int &x, int &y) {
	return detect_borders (field, x, y);
}

int temperature_in_front(int &x, int &y, int &dir) {
	return temperature_in_front (field, x, y, dir);
}

void evolve_field () {
	field.evolve_field ();
}

void print_field (int &robot_x, int &robot_y, int &dir) {
	field.print_field (robot_x, robot_y, dir);
}

